/* Copyright (c) 2013 Nordic Semiconductor. All Rights Reserved.
 *
 * Use of this source code is governed by a BSD-style license that can be
 * found in the license.txt file.
 */

/** @file
 * @brief RTC/PPI/ADC modules example
 *
 * This example demonstrates how to sample voltage with the ADC on an input pin.
 * RTC timer is enabled with a TICK frequency of 10Hz. Each RTC TICK triggers
 * ADC conversion. PPI channel connects the RTC TICK event to the ADC START task.
 * Whenever the ADC samples the voltage on the input pin, the result is displayed
 * on port 1, i.e. LEDs 0-7 (Pins P0.08-P0.15)
 */

#include <stdbool.h>
#include "nrf.h"
#include "nrf_gpio.h"
#include "SEGGER_RTT.h"
#include "boards.h"

#define LFCLK_FREQUENCY           (32768UL)                 /*!< LFCLK frequency in Hertz, constant */
#define RTC_FREQUENCY             (10UL)                     /*!< required RTC working clock RTC_FREQUENCY Hertz. Changable */
#define COUNTER_PRESCALER         ((LFCLK_FREQUENCY/RTC_FREQUENCY) - 1)  /* f = LFCLK/(prescaler + 1) */

#define ADC_RESOLUTION            ADC_CONFIG_RES_10bit

static uint8_t m_adc_pin_configuration;

/** Configures port 1 as output
 */
static void gpio_config(void)
{
	// Configure LED-pins as outputs.
	LEDS_CONFIGURE(LEDS_MASK);
}

static void hfclk_config(void)
{
	/* Start 16 MHz crystal oscillator */
  NRF_CLOCK->EVENTS_HFCLKSTARTED = 0;
  NRF_CLOCK->TASKS_HFCLKSTART = 1;

  /* Wait for the external oscillator to start up */
  while (NRF_CLOCK->EVENTS_HFCLKSTARTED == 0) {}
}

static void lfclk_config(void)
{
  NRF_CLOCK->LFCLKSRC = (CLOCK_LFCLKSRC_SRC_Xtal << CLOCK_LFCLKSRC_SRC_Pos);
  NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
  NRF_CLOCK->TASKS_LFCLKSTART = 1;
	
	// Wait for the low frequency clock to start
  while (NRF_CLOCK->EVENTS_LFCLKSTARTED == 0) {}
  NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
}

/** Configure and start RTC
 */
static void rtc_config(void)
{
  NRF_RTC0->PRESCALER = COUNTER_PRESCALER;                   // Set prescaler to a TICK of RTC_FREQUENCY
  NRF_RTC0->EVTENSET = RTC_EVTENSET_TICK_Msk;                // Enable TICK event	
	NRF_RTC0->TASKS_START = 1;                                 // Start RTC0
}

static void ppi_init(void)
{
  // Configure PPI channel 0 to start ADC task
  NRF_PPI->CH[0].EEP = (uint32_t)&NRF_RTC0->EVENTS_TICK;
  NRF_PPI->CH[0].TEP = (uint32_t)&NRF_ADC->TASKS_START;

  // Enable PPI channel 0
  NRF_PPI->CHEN = (PPI_CHEN_CH0_Enabled << PPI_CHEN_CH0_Pos);
}

void ADC_pin_configure(uint8_t pin_number)
{
	NRF_ADC->CONFIG	|= (pin_number << ADC_CONFIG_PSEL_Pos);
	m_adc_pin_configuration = pin_number;
}

/** Configures and enables the ADC
 */
void ADC_init(uint8_t pin_number)
{	
	/* Enable interrupt on ADC sample ready event*/		
	NRF_ADC->INTENSET = ADC_INTENSET_END_Msk;
	NVIC_EnableIRQ(ADC_IRQn);	
	
	// config ADC
	NRF_ADC->CONFIG	= (ADC_CONFIG_EXTREFSEL_None << ADC_CONFIG_EXTREFSEL_Pos) 								/* Bits 17..16 : ADC external reference pin selection. */								                  /*!< Use analog input 6 as analog input (P0.05). */
										| (ADC_CONFIG_REFSEL_VBG << ADC_CONFIG_REFSEL_Pos)											/*!< Use internal 1.2V bandgap voltage as reference for conversion. */
										| (ADC_CONFIG_INPSEL_AnalogInputOneThirdPrescaling << ADC_CONFIG_INPSEL_Pos) 	/*!< Analog input specified by PSEL with no prescaling used as input for the conversion. */
										| (ADC_CONFIG_RES_8bit << ADC_CONFIG_RES_Pos);		/*!< 8bit ADC resolution. */ 	
	ADC_pin_configure(pin_number);
	
	/* Enable ADC*/
	NRF_ADC->ENABLE = ADC_ENABLE_ENABLE_Enabled;	
}

void write_adc_result_to_leds(uint16_t adc_result)
{
  // Write 3 LSB ADC result to the LEDs on nRF51-Dongle (PCA10031) and nRF51-DK (PCA10028)
	nrf_gpio_pin_write(LED_1, (adc_result & 0x01));
	nrf_gpio_pin_write(LED_2, (adc_result & 0x02));
	nrf_gpio_pin_write(LED_3, (adc_result & 0x04));

	//If this is the nRF51DK (PCA10028) with 4 leds, write LED_4 also
	if (LEDS_NUMBER == 4) 
	{ 
		nrf_gpio_pin_write(LED_4, (NRF_ADC->RESULT & 0x08));
	}
}

void adc_calibrate(uint16_t adc_result, uint16_t * adc_result_calibrated, uint8_t adc_resolution)
{
	uint32_t uicr_value;
	uint8_t offset_error;
	uint8_t gain_error;
	
	// Apply ADC calibration described on 
	// https://devzone.nordicsemi.com/question/21653/how-to-calibrate-the-nrf51-adc-to-correct-offset-and-gain-error/
	uicr_value = *(uint32_t *)0x10000024;       //Read ADC gain and offset error from UICR
	offset_error = uicr_value;
	gain_error = uicr_value >> 8;
	
	if(adc_resolution == ADC_CONFIG_RES_10bit)
	{
		*adc_result_calibrated = NRF_ADC->RESULT * (1024 + gain_error) / 1024 + offset_error;    //calibrate
	}
	else
	{
		//Calibration method not valid for given ADC resolution. Return adc value uncalibrated.
		*adc_result_calibrated = adc_result;      
	}
}	

/* Interrupt handler for ADC data ready event */
void ADC_IRQHandler(void)
{
	uint16_t adc_result_calibrated;

	/* Clear dataready event */
  NRF_ADC->EVENTS_END = 0;	
	
	//Attempt to calibrate the ADC result. Only valid for nRF51 third revision IC
	adc_calibrate(NRF_ADC->RESULT, &adc_result_calibrated, ADC_RESOLUTION);
	
	switch(m_adc_pin_configuration)
	{
		case ADC_CONFIG_PSEL_AnalogInput5: 
			//Print ADC result on LEDs. Not very meaningful as only 4 bits are shown on nRF51DK
			write_adc_result_to_leds(adc_result_calibrated);

			//print ADC result on RTT, open J-Link RTT Viewer on port 0 to see the ADC result.
			SEGGER_RTT_printf(0, "ADC result AIN pin 5:  %d\n", adc_result_calibrated);	
			
			//Configure ADC for the second pin
			ADC_pin_configure(ADC_CONFIG_PSEL_AnalogInput6);
		
			//start sampling on a second pin
			NRF_ADC->TASKS_START = 1;
			break;
		
		case ADC_CONFIG_PSEL_AnalogInput6: 
			//print ADC result on RTT, open J-Link RTT Viewer on port 0 to see the ADC result.
			SEGGER_RTT_printf(0, "ADC result AIN pin 5:  %d\n", adc_result_calibrated);	
			
			//Configure ADC for the second pin
			ADC_pin_configure(ADC_CONFIG_PSEL_AnalogInput6);
			break;
	}
}


/**
 * main function
 */
int main(void)
{
  gpio_config();
	lfclk_config();
	hfclk_config();
  rtc_config();
	ppi_init();
	ADC_init(ADC_CONFIG_PSEL_AnalogInput5);
	
	SEGGER_RTT_WriteString(0, "Device starting ...\n");
  while (true)
  {	
		//Put CPU to sleep while waiting for interrupt
		__WFE();
		__SEV();
		__WFE();
  }
}

